/*  Copyright (C) 2020 Andrea G. Monaco
 *
 *  This file is part of libre-sapienza.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */


/* Code for speaking with the servers */


#include "config.h"

#include <stdio.h>
#include <string.h>

#include <curl/curl.h>
#include <jansson.h>

#include <libintl.h>
#define _(String) gettext (String)

#include "network.h"
#include "util.h"
#include "main.h"
#include "exam.h"
#include "student.h"


char url_buffer [URL_BUFFER_SIZE]; /* buffer used to prepare the url */

char *token; /* authentication token by the server */

CURL *curl_handle; /* libcurl object used for talking to the server */

char *response = NULL; /* buffer for the body of http responses */
size_t resp_s = 0;     /* size of the response buffer */
size_t resp_c = 0;     /* cursor of the next element to write in the buffer; response [resp_c] should always be \0 */
 


/* callback called by the curl library with the body of the http response.
   as specified by the library doc, it returns the number of processed bytes */
size_t
write_callback (char *ptr, size_t size, size_t nmemb, void *userdata)
{
  size_t realsize = size * nmemb;
  if (resp_c + realsize > resp_s)
    {
      response = realloc (response, resp_c + realsize);
    }
  if (!response)
    { 
      fprintf (stderr, _("memory exhausted!\n"));
      return 0;
    }
  resp_s = resp_c + realsize;
  memcpy (response + resp_c, ptr, realsize);
  resp_c += realsize;
  response [resp_c] = 0;
 
  return realsize;
}


/* empty the response buffer, without freeing any memory */
void
empty_response (void)
{
  resp_c = response [0] = 0;
}


/* make a login attempt with given id and password and return token */
void
attempt_login (const char *pass)
{
  struct curl_slist *headers;
  CURLcode ret;
  json_t *root, *esito, *flag_esito_j, *output;
  json_int_t flag_esito;
  json_error_t error;
  
  curl_easy_setopt (curl_handle, CURLOPT_URL, BASE_URL "/autenticazione");

  /* create a linked list of header fields; this is freed by the end of the function */
  headers = curl_slist_append (NULL, "Accept: application/json");
  headers = curl_slist_append (headers, "Content-EventType: application/x-www-form-urlencoded");
  curl_easy_setopt (curl_handle, CURLOPT_HTTPHEADER, headers);

  snprintf (url_buffer, URL_BUFFER_SIZE, "key=1nf0r1cc1&matricola=%d&stringaAutenticazione=%s", studentid, pass);
  curl_easy_setopt (curl_handle, CURLOPT_POSTFIELDS, url_buffer);

  /* prepare the response buffer */
  response = xmalloc (2048);
  resp_s = 2048;
  resp_c = 0;
  response [0] = 0;

  curl_easy_setopt (curl_handle, CURLOPT_WRITEFUNCTION, write_callback);
  
  ret = curl_easy_perform (curl_handle);
  
  if (ret != CURLE_OK)
    {
      fprintf (stderr, _("curl_easy_perform() failed: %s\n"),
	       curl_easy_strerror (ret));
      exit (1);
    }

  if (DEBUG)
    printf ("%s\n", response);

  root = json_loads (response, 0, &error);
  esito = json_object_get (root, "esito");
  flag_esito_j = json_object_get (esito, "flagEsito");
  flag_esito = json_integer_value (flag_esito_j);
  
  if (!root)
    {
      fprintf (stderr, _("error: on line %d: %s\n"), error.line, error.text);
      exit (1);
    }
  switch (flag_esito)
    {
    case -6:
      fprintf (stderr, _("invalid student id or password\n"));
      exit (1);
      break;
    case -4:
      fprintf (stderr, _("user not enabled for Infostud service\n"));
      exit (1);
      break;
    case -2:
      fprintf (stderr, _("password expired\n"));
      exit (1);
      break;
    case -1:
      fprintf (stderr, _("invalid credentials\n"));
      exit (1);
      break;
    case 0:
      printf (_("login successful!\n"));
      break;
    default:
      fprintf (stderr, _("i don't know what the hell is going on\n"));
      exit (1);
      break;
    }

  output = json_object_get (root, "output");
  token = xmalloc (64);
  strncpy (token, json_string_value (output), 64);

  /* decrement reference counts */
  json_decref (root);

  /* i'm not sure if these are needed, i don't fully understand jansson library memory allocation */
  /*  json_decref (esito);
      json_decref (flag_esito_j); 
      json_decref (output); */
  
  /* reset header and fields, plus free linked list of headers */
  curl_easy_setopt (curl_handle, CURLOPT_HTTPHEADER, NULL);
  curl_easy_setopt (curl_handle, CURLOPT_POSTFIELDS, NULL);
  curl_slist_free_all (headers);
}


/* request available exams to the server */
int
request_available_exams (void)
{
  json_error_t error;
  
  snprintf (url_buffer, URL_BUFFER_SIZE, BASE_URL "/studente/%d/insegnamentisostenibili?ingresso=%s", studentid, token);
  perform_get_request (url_buffer);
  
  available_exams = json_loads (response, 0, &error);
  if (!available_exams)
    {
      fprintf (stderr, _("error: on line %d: %s\n"), error.line, error.text);
      return 0;
    }
  return 1;
}


int
request_completed_exams (void)
{
  json_error_t error;
  
  snprintf (url_buffer, URL_BUFFER_SIZE, BASE_URL "/studente/%d/esamiall?ingresso=%s", studentid, token);    
  perform_get_request (url_buffer);
  
  completed_exams = json_loads (response, 0, &error);
  if (!completed_exams)
    {
      fprintf (stderr, _("error: on line %d: %s\n"), error.line, error.text);
      return 0;
    }
  return 1;
}


int
request_active_reservations (void)
{
  json_error_t error;

  snprintf (url_buffer, URL_BUFFER_SIZE, BASE_URL "/studente/%d/prenotazioni?ingresso=%s", studentid, token);
  perform_get_request (url_buffer);
  
  active_reservations = json_loads (response, 0, &error);
  if (!active_reservations)
    {
      fprintf (stderr, _("error: on line %d: %s\n"), error.line, error.text);
      return 0;
    }
  return 1;
}


int
request_student_info (void)
{
  json_error_t error;

  snprintf (url_buffer, URL_BUFFER_SIZE, BASE_URL "/studente/%d?ingresso=%s", studentid, token);
  perform_get_request (url_buffer);
  
  student_info = json_loads (response, 0, &error);
  if (!student_info)
    {
      fprintf (stderr, _("error: on line %d: %s\n"), error.line, error.text);
      return 0;
    }
  return 1;
}


int
request_student_photo (void)
{
  json_error_t error;
  
  snprintf (url_buffer, URL_BUFFER_SIZE, BASE_URL "/studente/%d/esamiall?ingresso=%s", studentid, token);    
  perform_get_request (url_buffer);
  
  completed_exams = json_loads (response, 0, &error);
  if (!completed_exams)
    {
      fprintf (stderr, _("error: on line %d: %s\n"), error.line, error.text);
      return 0;
    }
  return 1;
}


/* perform a HTTP GET request to the server at the given url. Most infostud functions are accessed this way */
int
perform_get_request (const char *url)
{
  CURLcode ret;

  empty_response ();
  curl_easy_setopt (curl_handle, CURLOPT_URL, url);
  curl_easy_setopt (curl_handle, CURLOPT_WRITEFUNCTION, write_callback);
  curl_easy_setopt (curl_handle, CURLOPT_HTTPGET, 1L);
  curl_easy_setopt (curl_handle, CURLOPT_HTTPHEADER, NULL);
  
  ret = curl_easy_perform (curl_handle);
  if (ret != CURLE_OK)
    {
      fprintf (stderr, _("curl_easy_perform() failed: %s\n"),
	       curl_easy_strerror (ret));
      return 0;
    }

  if (PRINT_JSON)
    {
      fprintf (stderr, "%s\n", response);
    }
  
  return 1;
}


